#!/usr/bin/perl

# Alain's Web counter
#
# Usage: <img src="count.cgi?W=width">
#   where
#       width is the number of digits to be displayed
#       id is a unique identifier, which can be used to manage several counters
#         with a same script
#       digits is the file name of a image which contains the digits from 0
#       to 9. Attention: each digit should have a contant width!
#
# Or:
#       <script src="count.cgi?js=1"></script>
#  Displays a text counter using javascript
# Combine both with:
#  <script src="count.cgi?id=label&js=1"></script>
#  <noscript><img src="count.cgi?W=width"></noscript>
# This displays a text counter where Javascript is available, and a graphical
# counter otherwise

use GD;
use CGI;
use CGI::Carp qw(fatalsToBrowser);
use strict;
use vars qw/$digits/;
use English;
use POSIX qw(strftime);

my $thisScript=$0;
my ($directory, $scriptName);
if ($thisScript =~ /(.*)\/(.*)/) {
  $directory=$1;
  $scriptName=$2;
} else {
  $directory=".";
  $scriptName=$thisScript;
}


my $query = new CGI;
$query->import_names('ARG');

my $httpTimeFmt = "%a, %d %b %G %T GMT";

my $counterModified;

my $ctrFile;
my $timestr;

sub statCtrFile {
  my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
      $atime,$mtime,$ctime,$blksize,$blocks) = stat($ctrFile);
  $counterModified = $mtime;
  $timestr = strftime $httpTimeFmt, gmtime($counterModified);
}

sub header {
  my ($query, $fileType, $testMode) = @_;
  if($testMode) {
    print $query->header(-type=>$fileType);
  } else {
    print $query->header(-type=>$fileType,
			 -Cache_control=>'private',
			 -last_modified=>$timestr,
			 -expires=>'+1d');
  }
}

################################################################
# Counter locking                                              #
################################################################
my $LOCK_SH = 1;
my $LOCK_EX = 2;
my $LOCK_NB = 4;
my $LOCK_UN = 8;

$ctrFile="$directory/counter.txt";

my $W=6;
my $n; # Value to display
my $N; # Value to write back to counter file

my $shouldWrite = 0;

my $modifSince = $ENV{HTTP_IF_MODIFIED_SINCE};
my $doIncrement;

my $canPng;

my $fileType;
if (defined $ARG::js) {
  $fileType = 'text/x-javascript';
} else {
  if ($ENV{"HTTP_ACCEPT"} =~ /image\/png/) {
    $canPng = 1;
  }
  if($canPng) {
    $fileType = 'image/png';
  } else {
    $fileType = 'image/jpeg';
  }
}

if (!defined $ARG::test) {
  statCtrFile;
  $doIncrement = 1;
  if(defined $modifSince) {
    if($modifSince >= $counterModified ) {
      # Not modified, return short answer
      print $query->header(-status=>'304 Not Modified',
			   -Cache_control=>'private',
			   -expires=>'+1d',
			   -last_modified=>$timestr);
      exit(0);
    }
    $doIncrement = 0;
  }
  if($ENV{REQUEST_METHOD} eq "HEAD") {
    # Head request only, return short answer
    print $query->header(-type=>$fileType,
			 -Cache_control=>'private',
			 -expires=>'+1d',
			 -last_modified=>$timestr);
    exit(0);
  }
  my $ua = $ENV{HTTP_USER_AGENT};
  if(defined $ua &&
     (($ua =~ /^Mozilla.4/ &&
       $ua !~ /compatible/) ||
      ($ua =~ /Konqueror/)) &&
     $ENV{HTTP_PRAGMA} eq "no-cache") {
    # Netscape 4 or Konqueror doing a reload. These browsers are not capable
    # of using IF_MODIFIED_SINCE properly... Even IE does it correctly!
    $doIncrement = 0;
  }
} else {
  # test mode, do not increment
  $doIncrement = 0;
}

if (defined $ARG::test) {
  $W=10;
  $n = 123456789;
} else {
  if (open(ID, "<$ctrFile")) {
    if($doIncrement) {
      # Only increment counter if first access...

      flock ID, $LOCK_EX;
      $n = <ID>;
      $n++;
    } else {
      $n = <ID>;
    }
    # Do not close <ID>, we want to conserve the lock...
  } else {
    $doIncrement = 1;
    $n=1;
  }
  $N = $n;

  if ($doIncrement) {
    open(OUT, ">$ctrFile.$$");
    print OUT "$N\n";
    close(OUT);
    rename ("$ctrFile.$$", "$ctrFile");
    statCtrFile;
  }
}

my $dummy=$ARG::js; # Shut up warning about ARG::js only being used once...
if (defined $ARG::js) {
  header($query, 'text/x-javascript', 0);
  print "document.write(\"$N\")\n";
} else {
  if (defined($ARG::W)) {
    $W = $ARG::W;
  }

  if (!defined $digits) {
    $digits = "$directory/digits.jpg";
    if (! -f $digits) {
      $digits = "/usr/local/lib/webcounter/digits/simple.jpg";
    }
  }

  my ($maxWidth, $maxHeight);
  my @digitImg;
  my @sxoffset;
  my @syoffset;
  my @width;
  my @height;

  if (-f "$digits") {
    open(JPG, "$digits") || 
      die "Could not open digitfile $digits ($ERRNO)\n";
    my $digitImg = newFromJpeg GD::Image('JPG') ||
      die "Could not read digitfile as jgp\n";
    close(JPG);
    ($maxWidth, $maxHeight) = $digitImg->getBounds();
    $maxWidth = $maxWidth / 10;
    for(my $d=0; $d<10; $d++) {
      $digitImg[$d] = $digitImg;
      ($sxoffset[$d], $syoffset[$d]) = ($maxWidth*$d, 0);
      ($width[$d], $height[$d]) = ($maxWidth, $maxHeight);
    }

    my $digitsRoot=$digits;
    $digitsRoot =~ s/\.[^.]*$//g;
    my $d = 0;
    if(-f "$digitsRoot.pos") {
      open(WIDTH, "$digitsRoot.pos") || 
	die "Could not open digit positions file ($ERRNO)\n";
      while(<WIDTH>) {
	chop;
	$sxoffset[$d] = $_;
	if ($d > 0) {
	  $width[$d-1] = $sxoffset[$d] - $sxoffset[$d-1];
	}
	$d++;
      }
      close(WIDTH);
      if ($d < 11) {
	die "Not enough data $d\n";
      }
    }

  } elsif (-d $digits) {
    ($maxWidth, $maxHeight) = ( 0, 0);
    for(my $d=0; $d < 10; $d++) {
      open(JPG, "$digits/$d.jpg") || die;
      $digitImg[$d] = newFromJpeg GD::Image('JPG') || die;
      ($width[$d], $height[$d]) = $digitImg[$d]->getBounds();
      ($sxoffset[$d], $syoffset[$d]) = (0, 0);
      close(JPG);
    }
  } else {
    die "Digits $digits not found\n";
  }

  # calculate the max width and height
  ($maxWidth, $maxHeight) = (0,0);
  for(my $d=0; $d < 10; $d++) {
      if ($width[$d] > $maxWidth) {
	$maxWidth = $width[$d];
      }
      if ($height[$d] > $maxHeight) {
	$maxHeight = $height[$d];
      }
  }

  my @txoffset;
  my @tyoffset;

  # calculate the "target" offsets
  for(my $d=0; $d < 10; $d++) {
    $txoffset[$d] = ($maxWidth - $width[$d]) / 2;
    $tyoffset[$d] = ($maxHeight - $height[$d]) / 2;
  }

  my $testMode;
  my $imWidth;
  if (defined $ARG::test) {
    my $twidth=1;
    for(my $d=0; $d < 10; $d++) {
      $twidth+= $width[$d]+1;
    }
    $testMode = 1;
    $imWidth = $twidth;
  } else {
    $testMode = 0;
    $imWidth = $maxWidth*$W+2;
  }

  my $im = new GD::Image($imWidth, $maxHeight+2);
  my $white = $im->colorAllocate(255,255,255);
  my $black = $im->colorAllocate(0,0,0);
  $im->rectangle(0,0, $imWidth-1, $maxHeight+1, $black);

  if(defined $ARG::test) {
    my $pos=0;
    for(my $d=0; $d < 10; $d++) {
      $im->line($pos,0, $pos, $maxHeight+1, $black);
      $pos++;
      $im->copy($digitImg[$d],
		$pos, 1, 
		$sxoffset[$d], $syoffset[$d], 
		$width[$d], $height[$d]);
      $pos += $width[$d];
    }
  } else {
    for(my $i=0; $i < $W; $i++) {
      my $d = $n % 10;
      $im->copy($digitImg[$d],
		1+($W-1-$i)*$maxWidth + $txoffset[$d], 1 + $tyoffset[$d],
		$sxoffset[$d], $syoffset[$d],
		$width[$d], $height[$d]);
      $n = int($n / 10);
    }
  }

#  $im->interlaced('true');

#  print "Content-type: text/plain\n\n";
#  foreach $key (keys %ENV) {
#    print "$key=$ENV{$key}\n";
#  }
#  exit(0);


  binmode STDOUT;
  header($query,$fileType, $testMode);
  if ($canPng) {
    print $im->png;
  } else {
    print $im->jpeg;
  }
}

1;
